"""
OpenLIT Instrumentors
"""

import importlib

# Mapping of instrumentor names to their required Python packages
MODULE_NAME_MAP = {
    "openai": "openai",
    "anthropic": "anthropic",
    "cohere": "cohere",
    "mistral": "mistralai",
    "bedrock": "boto3",
    "vertexai": "vertexai",
    "groq": "groq",
    "ollama": "ollama",
    "gpt4all": "gpt4all",
    "elevenlabs": "elevenlabs",
    "vllm": "vllm",
    "google-ai-studio": "google.genai",
    "azure-ai-inference": "azure.ai.inference",
    "langchain": "langchain",
    "langchain_community": "langchain_community",
    "llama_index": "llama_index",
    "haystack": "haystack",
    "mem0": "mem0",
    "chroma": "chromadb",
    "pinecone": "pinecone",
    "qdrant": "qdrant_client",
    "milvus": "pymilvus",
    "transformers": "transformers",
    "litellm": "litellm",
    "crewai": "crewai",
    "ag2": "ag2",
    "autogen": "autogen",
    "pyautogen": "pyautogen",
    "multion": "multion",
    "dynamiq": "dynamiq",
    "agno": "agno",
    "reka-api": "reka",
    "premai": "premai",
    "julep": "julep",
    "astra": "astrapy",
    "ai21": "ai21",
    "controlflow": "controlflow",
    "assemblyai": "assemblyai",
    "crawl4ai": "crawl4ai",
    "firecrawl": "firecrawl",
    "letta": "letta_client",
    "together": "together",
    "openai-agents": "agents",
    "pydantic_ai": "pydantic_ai",
    "sarvam": "sarvamai",
}

# Dictionary mapping instrumentor names to their full module paths
INSTRUMENTOR_MAP = {
    "openai": "openlit.instrumentation.openai.OpenAIInstrumentor",
    "anthropic": "openlit.instrumentation.anthropic.AnthropicInstrumentor",
    "cohere": "openlit.instrumentation.cohere.CohereInstrumentor",
    "mistral": "openlit.instrumentation.mistral.MistralInstrumentor",
    "bedrock": "openlit.instrumentation.bedrock.BedrockInstrumentor",
    "vertexai": "openlit.instrumentation.vertexai.VertexAIInstrumentor",
    "groq": "openlit.instrumentation.groq.GroqInstrumentor",
    "ollama": "openlit.instrumentation.ollama.OllamaInstrumentor",
    "gpt4all": "openlit.instrumentation.gpt4all.GPT4AllInstrumentor",
    "elevenlabs": "openlit.instrumentation.elevenlabs.ElevenLabsInstrumentor",
    "vllm": "openlit.instrumentation.vllm.VLLMInstrumentor",
    "google-ai-studio": "openlit.instrumentation.google_ai_studio.GoogleAIStudioInstrumentor",
    "azure-ai-inference": "openlit.instrumentation.azure_ai_inference.AzureAIInferenceInstrumentor",
    "langchain": "openlit.instrumentation.langchain.LangChainInstrumentor",
    "langchain_community": "openlit.instrumentation.langchain_community.LangChainCommunityInstrumentor",
    "llama_index": "openlit.instrumentation.llamaindex.LlamaIndexInstrumentor",
    "haystack": "openlit.instrumentation.haystack.HaystackInstrumentor",
    "mem0": "openlit.instrumentation.mem0.Mem0Instrumentor",
    "chroma": "openlit.instrumentation.chroma.ChromaInstrumentor",
    "pinecone": "openlit.instrumentation.pinecone.PineconeInstrumentor",
    "qdrant": "openlit.instrumentation.qdrant.QdrantInstrumentor",
    "milvus": "openlit.instrumentation.milvus.MilvusInstrumentor",
    "transformers": "openlit.instrumentation.transformers.TransformersInstrumentor",
    "litellm": "openlit.instrumentation.litellm.LiteLLMInstrumentor",
    "crewai": "openlit.instrumentation.crewai.CrewAIInstrumentor",
    "ag2": "openlit.instrumentation.ag2.AG2Instrumentor",
    "multion": "openlit.instrumentation.multion.MultiOnInstrumentor",
    "autogen": "openlit.instrumentation.ag2.AG2Instrumentor",
    "pyautogen": "openlit.instrumentation.ag2.AG2Instrumentor",
    "dynamiq": "openlit.instrumentation.dynamiq.DynamiqInstrumentor",
    "agno": "openlit.instrumentation.agno.AgnoInstrumentor",
    "reka-api": "openlit.instrumentation.reka.RekaInstrumentor",
    "premai": "openlit.instrumentation.premai.PremAIInstrumentor",
    "julep": "openlit.instrumentation.julep.JulepInstrumentor",
    "astra": "openlit.instrumentation.astra.AstraInstrumentor",
    "ai21": "openlit.instrumentation.ai21.AI21Instrumentor",
    "controlflow": "openlit.instrumentation.controlflow.ControlFlowInstrumentor",
    "assemblyai": "openlit.instrumentation.assemblyai.AssemblyAIInstrumentor",
    "crawl4ai": "openlit.instrumentation.crawl4ai.Crawl4AIInstrumentor",
    "firecrawl": "openlit.instrumentation.firecrawl.FireCrawlInstrumentor",
    "letta": "openlit.instrumentation.letta.LettaInstrumentor",
    "together": "openlit.instrumentation.together.TogetherInstrumentor",
    "openai-agents": "openlit.instrumentation.openai_agents.OpenAIAgentsInstrumentor",
    "pydantic_ai": "openlit.instrumentation.pydantic_ai.PydanticAIInstrumentor",
    "sarvam": "openlit.instrumentation.sarvam.SarvamInstrumentor",
}


def get_instrumentor_class(name):
    """
    Get instrumentor class by name.

    Args:
        name (str): Name of the instrumentor

    Returns:
        class: Instrumentor class or None if not found
    """
    if name not in INSTRUMENTOR_MAP:
        return None

    module_path, class_name = INSTRUMENTOR_MAP[name].rsplit(".", 1)

    try:
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError):
        return None


def get_all_instrumentors():
    """
    Get all available instrumentor instances.

    Returns:
        dict: Dictionary of instrumentor instances
    """
    instances = {}

    for name in INSTRUMENTOR_MAP:
        instrumentor_class = get_instrumentor_class(name)
        if instrumentor_class:
            try:
                instances[name] = instrumentor_class()
            except Exception:
                pass  # Skip instrumentors that fail to instantiate

    return instances



"""
OpenLIT OpenAI Agents Instrumentation - Native TracingProcessor Implementation
"""

import time
from typing import Any, Dict, TYPE_CHECKING

from opentelemetry.trace import SpanKind, Status, StatusCode, set_span_in_context

from openlit.__helpers import (
    common_framework_span_attributes,
    handle_exception,
    get_chat_model_cost,
)
from openlit.semcov import SemanticConvention

# Try to import agents framework components with fallback
try:
    from agents import TracingProcessor

    if TYPE_CHECKING:
        from agents import Trace, Span
    TRACING_AVAILABLE = True
except ImportError:
    # Create dummy class for when agents is not available
    class TracingProcessor:
        """Dummy TracingProcessor class for when agents is not available"""

        def force_flush(self):
            """Dummy force_flush method"""
            return None

        def shutdown(self):
            """Dummy shutdown method"""
            return None

    if TYPE_CHECKING:
        # Type hints only - these don't exist at runtime when agents unavailable
        Trace = Any
        Span = Any

    TRACING_AVAILABLE = False


class OpenLITTracingProcessor(TracingProcessor):
    """
    OpenAI Agents tracing processor that integrates with OpenLIT observability.

    This processor enhances OpenAI Agents' native tracing system with OpenLIT's
    comprehensive observability features including business intelligence,
    cost tracking, and performance metrics.
    """

    def __init__(
        self,
        tracer,
        version,
        environment,
        application_name,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
        detailed_tracing,
        **kwargs,
    ):
        """Initialize the OpenLIT tracing processor."""
        super().__init__()

        # Core configuration
        self.tracer = tracer
        self.version = version
        self.environment = environment
        self.application_name = application_name
        self.pricing_info = pricing_info
        self.capture_message_content = capture_message_content
        self.metrics = metrics
        self.disable_metrics = disable_metrics
        self.detailed_tracing = detailed_tracing

        # Internal tracking
        self.active_spans = {}
        self.span_stack = []

    def start_trace(self, trace_id: str, name: str, **kwargs):
        """
        Start a new trace with OpenLIT enhancements.

        Args:
            trace_id: Unique trace identifier
            name: Trace name
            **kwargs: Additional trace metadata
        """
        try:
            # Generate span name using OpenTelemetry conventions
            span_name = self._get_span_name(name, **kwargs)

            # Start root span with OpenLIT context
            span = self.tracer.start_as_current_span(
                span_name,
                kind=SpanKind.CLIENT,
                attributes={
                    SemanticConvention.GEN_AI_SYSTEM: "openai_agents",
                    SemanticConvention.GEN_AI_OPERATION: SemanticConvention.GEN_AI_OPERATION_TYPE_WORKFLOW,
                    "trace.id": trace_id,
                    "trace.name": name,
                },
            )

            # Create scope for common attributes
            scope = type("GenericScope", (), {})()
            scope._span = span  # pylint: disable=protected-access
            scope._start_time = time.time()  # pylint: disable=protected-access
            scope._end_time = None  # pylint: disable=protected-access

            # Apply common framework attributes
            common_framework_span_attributes(
                scope,
                "openai_agents",
                "api.openai.com",
                443,
                self.environment,
                self.application_name,
                self.version,
                name,
            )

            # Track active span
            self.active_spans[trace_id] = span
            self.span_stack.append(span)

            return span

        except Exception as e:  # pylint: disable=broad-exception-caught
            # Graceful degradation
            handle_exception(None, e)
            return None

    def end_trace(self, trace_id: str, **kwargs):
        """
        End an active trace.

        Args:
            trace_id: Trace identifier to end
            **kwargs: Additional metadata
        """
        try:
            span = self.active_spans.get(trace_id)
            if span:
                # Set final attributes and status
                span.set_status(Status(StatusCode.OK))

                # End span
                span.end()

                # Cleanup tracking
                if trace_id in self.active_spans:
                    del self.active_spans[trace_id]
                if span in self.span_stack:
                    self.span_stack.remove(span)

        except Exception as e:  # pylint: disable=broad-exception-caught
            handle_exception(span if span else None, e)

    def _get_span_name(self, operation_name: str, **metadata) -> str:
        """
        Generate OpenTelemetry-compliant span names.

        Args:
            operation_name: Base operation name
            **metadata: Additional context for naming

        Returns:
            Formatted span name following semantic conventions
        """
        # Extract context for naming
        agent_name = metadata.get("agent_name", "")
        model_name = metadata.get("model_name", "")
        tool_name = metadata.get("tool_name", "")
        workflow_name = metadata.get("workflow_name", "")

        # Apply OpenTelemetry semantic conventions for GenAI agents
        if "agent" in operation_name.lower():
            if agent_name:
                return f"invoke_agent {agent_name}"
            return "invoke_agent"
        if "chat" in operation_name.lower():
            if model_name:
                return f"chat {model_name}"
            return "chat response"
        if "tool" in operation_name.lower():
            if tool_name:
                return f"execute_tool {tool_name}"
            return "execute_tool"
        if "handoff" in operation_name.lower():
            target_agent = metadata.get("target_agent", "unknown")
            return f"invoke_agent {target_agent}"
        if "workflow" in operation_name.lower():
            if workflow_name:
                return f"workflow {workflow_name}"
            return "workflow"

        # Default case
        return operation_name

    def span_start(self, span_data, trace_id: str):
        """
        Handle span start events from OpenAI Agents.

        Args:
            span_data: Span data from agents framework
            trace_id: Associated trace identifier
        """
        try:
            # Extract span information
            span_name = getattr(span_data, "name", "unknown_operation")
            span_type = getattr(span_data, "type", "unknown")

            # Generate enhanced span name
            enhanced_name = self._get_span_name(
                span_name,
                agent_name=getattr(span_data, "agent_name", None),
                model_name=getattr(span_data, "model_name", None),
                tool_name=getattr(span_data, "tool_name", None),
            )

            # Determine span operation type
            operation_type = self._get_operation_type(span_type, span_name)

            # Start span with proper context
            parent_span = self.span_stack[-1] if self.span_stack else None
            context = set_span_in_context(parent_span) if parent_span else None

            span = self.tracer.start_as_current_span(
                enhanced_name,
                kind=SpanKind.CLIENT,
                context=context,
                attributes={
                    SemanticConvention.GEN_AI_SYSTEM: "openai_agents",
                    SemanticConvention.GEN_AI_OPERATION: operation_type,
                    "span.type": span_type,
                    "span.id": getattr(span_data, "span_id", ""),
                },
            )

            # Process specific span types
            self._process_span_attributes(span, span_data, span_type)

            # Track span
            span_id = getattr(span_data, "span_id", len(self.span_stack))
            self.active_spans[f"{trace_id}:{span_id}"] = span
            self.span_stack.append(span)

        except Exception as e:  # pylint: disable=broad-exception-caught
            handle_exception(None, e)

    def _get_operation_type(self, span_type: str, span_name: str) -> str:
        """Get operation type based on span characteristics."""
        type_mapping = {
            "agent": SemanticConvention.GEN_AI_OPERATION_TYPE_AGENT,
            "generation": SemanticConvention.GEN_AI_OPERATION_CHAT,
            "function": SemanticConvention.GEN_AI_OPERATION_CHAT,
            "tool": SemanticConvention.GEN_AI_OPERATION_CHAT,
            "handoff": SemanticConvention.GEN_AI_OPERATION_TYPE_AGENT,
        }

        # Check span type first
        for key, operation in type_mapping.items():
            if key in span_type.lower():
                return operation

        # Check span name
        for key, operation in type_mapping.items():
            if key in span_name.lower():
                return operation

        return SemanticConvention.GEN_AI_OPERATION_CHAT

    def _process_span_attributes(self, span, span_data, span_type: str):
        """Process and set span attributes based on span type."""
        try:
            # Common attributes
            if hasattr(span_data, "agent_name"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_AGENT_NAME, span_data.agent_name
                )

            if hasattr(span_data, "model_name"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_REQUEST_MODEL, span_data.model_name
                )

            # Agent-specific attributes
            if span_type == "agent":
                self._process_agent_span(span, span_data)

            # Generation-specific attributes
            elif span_type == "generation":
                self._process_generation_span(span, span_data)

            # Function/Tool-specific attributes
            elif span_type in ["function", "tool"]:
                self._process_function_span(span, span_data)

            # Handoff-specific attributes
            elif span_type == "handoff":
                self._process_handoff_span(span, span_data)

        except Exception as e:  # pylint: disable=broad-exception-caught
            handle_exception(span, e)

    def _process_agent_span(self, span, agent_span):
        """Process agent span data (unused parameter)."""
        # Agent-specific processing
        if hasattr(agent_span, "instructions"):
            span.set_attribute(
                SemanticConvention.GEN_AI_AGENT_DESCRIPTION,
                str(agent_span.instructions)[:500],
            )

        if hasattr(agent_span, "model"):
            span.set_attribute(
                SemanticConvention.GEN_AI_REQUEST_MODEL, agent_span.model
            )

    def _process_generation_span(self, span, generation_span):
        """Process generation span data."""
        # Set generation-specific attributes
        if hasattr(generation_span, "prompt"):
            span.set_attribute(
                SemanticConvention.GEN_AI_PROMPT, str(generation_span.prompt)[:1000]
            )

        if hasattr(generation_span, "completion"):
            span.set_attribute(
                SemanticConvention.GEN_AI_COMPLETION,
                str(generation_span.completion)[:1000],
            )

        if hasattr(generation_span, "usage"):
            usage = generation_span.usage
            if hasattr(usage, "prompt_tokens"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_USAGE_PROMPT_TOKENS, usage.prompt_tokens
                )
            if hasattr(usage, "completion_tokens"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_USAGE_COMPLETION_TOKENS,
                    usage.completion_tokens,
                )

    def _process_function_span(self, span, function_span):
        """Process function/tool span data."""
        if hasattr(function_span, "function_name"):
            span.set_attribute(
                SemanticConvention.GEN_AI_TOOL_NAME, function_span.function_name
            )

        if hasattr(function_span, "arguments"):
            span.set_attribute(
                "gen_ai.tool.arguments", str(function_span.arguments)[:500]
            )

        if hasattr(function_span, "result"):
            span.set_attribute("gen_ai.tool.result", str(function_span.result)[:500])

    def _process_handoff_span(self, span, handoff_span):
        """Process handoff span data."""
        if hasattr(handoff_span, "target_agent"):
            span.set_attribute("gen_ai.handoff.target_agent", handoff_span.target_agent)

        if hasattr(handoff_span, "reason"):
            span.set_attribute("gen_ai.handoff.reason", str(handoff_span.reason)[:200])

    def span_end(self, span_data, trace_id: str):
        """Handle span end events."""
        try:
            span_id = getattr(span_data, "span_id", "")
            span_key = f"{trace_id}:{span_id}"

            span = self.active_spans.get(span_key)
            if span:
                # Set final status
                if hasattr(span_data, "error") and span_data.error:
                    span.set_status(Status(StatusCode.ERROR, str(span_data.error)))
                else:
                    span.set_status(Status(StatusCode.OK))

                # End span
                span.end()

                # Cleanup
                if span_key in self.active_spans:
                    del self.active_spans[span_key]
                if span in self.span_stack:
                    self.span_stack.remove(span)

        except Exception as e:  # pylint: disable=broad-exception-caught
            handle_exception(span if "span" in locals() else None, e)

    def force_flush(self):
        """Force flush all pending spans."""
        try:
            # End any remaining spans
            for span in list(self.active_spans.values()):
                span.end()

            self.active_spans.clear()
            self.span_stack.clear()

        except Exception as e:  # pylint: disable=broad-exception-caught
            handle_exception(None, e)

    def shutdown(self):
        """Shutdown the processor."""
        self.force_flush()

    def _extract_model_info(self, span_data) -> Dict[str, Any]:
        """Extract model information from span data."""
        model_info = {}

        if hasattr(span_data, "model"):
            model_info["model"] = span_data.model
        if hasattr(span_data, "model_name"):
            model_info["model"] = span_data.model_name

        return model_info

    def _calculate_cost(
        self, model: str, prompt_tokens: int, completion_tokens: int
    ) -> float:
        """Calculate cost based on token usage."""
        try:
            return get_chat_model_cost(
                model, self.pricing_info, prompt_tokens, completion_tokens
            )
        except Exception:  # pylint: disable=broad-exception-caught
            return 0.0

    # Abstract method implementations required by OpenAI Agents framework
    def on_trace_start(self, trace):
        """Called when a trace starts - required by OpenAI Agents framework"""
        try:
            self.start_trace(
                getattr(trace, "trace_id", "unknown"),
                getattr(trace, "name", "workflow"),
            )
        except Exception:  # pylint: disable=broad-exception-caught
            pass

    def on_trace_end(self, trace):
        """Called when a trace ends - required by OpenAI Agents framework"""
        try:
            self.end_trace(getattr(trace, "trace_id", "unknown"))
        except Exception:  # pylint: disable=broad-exception-caught
            pass

    def on_span_start(self, span):
        """Called when a span starts - required by OpenAI Agents framework"""
        try:
            trace_id = getattr(span, "trace_id", "unknown")
            self.span_start(span, trace_id)
        except Exception:  # pylint: disable=broad-exception-caught
            pass

    def on_span_end(self, span):
        """Called when a span ends - required by OpenAI Agents framework"""
        try:
            trace_id = getattr(span, "trace_id", "unknown")
            self.span_end(span, trace_id)
        except Exception:  # pylint: disable=broad-exception-caught
            pass



"""
Module for monitoring async Pydantic AI API calls.
"""

from openlit.instrumentation.pydantic_ai.utils import (
    common_agent_run_async,
)


def async_agent_run(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for async GenAI function call
    """

    async def wrapper(wrapped, instance, args, kwargs):
        return await common_agent_run_async(
            wrapped,
            instance,
            args,
            kwargs,
            tracer,
            version,
            environment,
            application_name,
            capture_message_content,
            pricing_info=pricing_info,
        )

    return wrapper



# pylint: disable=duplicate-code, broad-exception-caught, too-many-statements, unused-argument, too-many-branches
"""
Module for monitoring async Letta calls following OpenTelemetry patterns.
"""

import logging
import asyncio
import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import handle_exception
from .utils import (
    get_span_name,
    process_letta_response,
    OPERATION_TYPE_MAP,
)

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)


def create_async_letta_wrapper(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Creates a unified telemetry wrapper for async Letta operations following OpenTelemetry patterns.

    This follows the same pattern as sync LiteLLM and CrewAI instrumentations but for async operations.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the async API call to add telemetry following OpenTelemetry semantic conventions.
        """

        async def async_wrapper(*args, **kwargs):
            # Extract operation type from endpoint
            endpoint_parts = gen_ai_endpoint.split(".")
            operation = endpoint_parts[-1] if endpoint_parts else "unknown"
            operation_type = OPERATION_TYPE_MAP.get(operation, "workflow")

            # Generate proper span name with context
            span_name = get_span_name(operation_type, gen_ai_endpoint, instance, kwargs)
            start_time = time.time()

            # Check if this is a streaming operation
            streaming = operation_type == "chat"  # Assume chat operations might stream

            if streaming:
                # Execute the operation first to check if it's actually streaming
                if asyncio.iscoroutinefunction(wrapped):
                    response = await wrapped(*args, **kwargs)
                else:
                    response = wrapped(*args, **kwargs)

                # Check if response is actually a stream
                if hasattr(response, "__aiter__") and hasattr(response, "__anext__"):
                    # For async streaming: create span without context manager (like LiteLLM)
                    span = tracer.start_span(span_name, kind=SpanKind.CLIENT)
                    # Note: For async streams, we'd need an async version of TracedLettaStream
                    # For now, we'll treat them as non-streaming
                    streaming = False

            # Non-streaming or non-chat operations
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                try:
                    if not streaming:
                        # Execute async operation
                        if asyncio.iscoroutinefunction(wrapped):
                            response = await wrapped(*args, **kwargs)
                        else:
                            response = wrapped(*args, **kwargs)

                    # Process response using common helpers
                    process_letta_response(
                        span,
                        response,
                        kwargs,
                        operation_type,
                        instance,
                        start_time,
                        environment,
                        application_name,
                        version,
                        gen_ai_endpoint,
                        capture_message_content,
                        pricing_info,
                    )

                    return response

                except Exception as e:
                    handle_exception(span, e)
                    logger.error("Error in async Letta trace creation: %s", e)
                    # Re-execute without instrumentation on error
                    if asyncio.iscoroutinefunction(wrapped):
                        return await wrapped(*args, **kwargs)
                    else:
                        return wrapped(*args, **kwargs)

        return async_wrapper(*args, **kwargs)

    return wrapper


# Async agent operations
def async_create_agent(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async agent operations wrapper (create, retrieve, modify, delete, list)"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Async message operations
def async_send_message(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async message operations wrapper (create, create_stream, list, modify, etc.)"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Async memory operations
def async_memory_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async memory operations wrapper (core memory, blocks)"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Async tool operations
def async_tool_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async tool operations wrapper (list, attach, detach, create)"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Async context operations
def async_context_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async context operations wrapper"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Async source operations
def async_source_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Async source operations wrapper"""
    return create_async_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )



"""Initializer of Auto Instrumentation of Agno Framework Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.agno.agno import (
    agent_run_wrap,
    agent_continue_run_wrap,
    agent_run_tool_wrap,
    agent_add_tool_wrap,
    session_memory_wrap,
    memory_add_wrap,
    memory_search_wrap,
    function_execute_wrap,
    reasoning_wrap,
    vectordb_search_wrap,
    vectordb_upsert_wrap,
    knowledge_search_wrap,
    knowledge_add_wrap,
    workflow_run_wrap,
    team_run_wrap,
    model_run_function_call_wrap,
    function_entrypoint_wrap,
    memory_operation_wrap,
    parallel_execution_wrap,
    reasoning_tool_wrap,
)
from openlit.instrumentation.agno.async_agno import (
    async_agent_run_wrap,
    async_agent_continue_run_wrap,
    async_function_execute_wrap,
    async_reasoning_wrap,
    async_workflow_run_wrap,
    async_team_run_wrap,
    async_model_run_function_call_wrap,
    async_function_entrypoint_wrap,
)

_instruments = ("agno >= 0.6.0",)

WORKFLOW_OPERATIONS = [
    # Core Agent Execution Methods
    "Agent.run",
    "Agent.continue_run",
    # Team Operations
    "Agent.create",
]

COMPONENT_OPERATIONS = [
    # Tool Management
    "Agent.add_tool",
    "Agent.set_tools",
    # Memory Operations
    "Agent.get_session_summary",
    "Agent.get_user_memories",
]


class AgnoInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Agno Framework's agent library.
    """

    def __init__(self):
        """Initialize the AgnoInstrumentor."""
        super().__init__()
        self._original_thread_pool_submit = None

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default_application")
        environment = kwargs.get("environment", "default_environment")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        disable_metrics = kwargs.get("disable_metrics")
        detailed_tracing = kwargs.get("detailed_tracing", False)

        # Get version
        try:
            version = importlib.metadata.version("agno")
        except importlib.metadata.PackageNotFoundError:
            version = "unknown"

        # CRITICAL: Patch ThreadPoolExecutor globally to preserve OpenTelemetry context
        self._patch_thread_pool_executor()

        # Workflow Operations (Always Instrumented)
        wrap_function_wrapper(
            "agno.agent.agent",
            "Agent.run",
            agent_run_wrap(
                "agent.run",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        wrap_function_wrapper(
            "agno.agent.agent",
            "Agent.continue_run",
            agent_continue_run_wrap(
                "agent.continue_run",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # CRITICAL: Agent._run_tool is the bridge between agent and tool execution
        wrap_function_wrapper(
            "agno.agent.agent",
            "Agent._run_tool",
            agent_run_tool_wrap(
                "agent._run_tool",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # Async Operations
        wrap_function_wrapper(
            "agno.agent.agent",
            "Agent.arun",
            async_agent_run_wrap(
                "agent.arun",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        wrap_function_wrapper(
            "agno.agent.agent",
            "Agent.acontinue_run",
            async_agent_continue_run_wrap(
                "agent.acontinue_run",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # Component Operations (Detailed Tracing Only) - Updated to match actual agno structure
        if detailed_tracing:
            # CRITICAL: Bridge span context between agent and tool execution
            wrap_function_wrapper(
                "agno.models.base",
                "Model.run_function_call",
                model_run_function_call_wrap(
                    "model.run_function_call",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # CRITICAL: Bridge span context for async function calls
            wrap_function_wrapper(
                "agno.models.base",
                "Model.arun_function_call",
                async_model_run_function_call_wrap(
                    "model.arun_function_call",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # CRITICAL: Also instrument the batch function calls method
            wrap_function_wrapper(
                "agno.models.base",
                "Model.run_function_calls",
                model_run_function_call_wrap(
                    "model.run_function_calls",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.models.base",
                "Model.arun_function_calls",
                async_model_run_function_call_wrap(
                    "model.arun_function_calls",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # CRITICAL: Memory operations that bypass model bridge
            wrap_function_wrapper(
                "agno.memory.v2.memory",
                "Memory.add_user_memory",
                memory_operation_wrap(
                    "memory.add_user_memory",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.memory.v2.memory",
                "Memory.search_user_memories",
                memory_operation_wrap(
                    "memory.search_user_memories",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Memory Operations (v2)
            wrap_function_wrapper(
                "agno.memory.v2.memory",
                "Memory.add_user_memory",
                memory_add_wrap(
                    "memory.add_user_memory",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.memory.v2.memory",
                "Memory.search_user_memories",
                memory_search_wrap(
                    "memory.search_user_memories",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Note: Memory class does not have async_add_user_memory or async_search_user_memories

            # VectorDB Operations (base class)
            wrap_function_wrapper(
                "agno.vectordb.base",
                "VectorDb.search",
                vectordb_search_wrap(
                    "vectordb.search",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.vectordb.base",
                "VectorDb.upsert",
                vectordb_upsert_wrap(
                    "vectordb.upsert",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Knowledge Operations
            wrap_function_wrapper(
                "agno.knowledge.agent",
                "AgentKnowledge.search",
                knowledge_search_wrap(
                    "knowledge.search",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.knowledge.agent",
                "AgentKnowledge.load",
                knowledge_add_wrap(
                    "knowledge.load",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.knowledge.agent",
                "AgentKnowledge.add_document_to_knowledge_base",
                knowledge_add_wrap(
                    "knowledge.add_document_to_knowledge_base",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Workflow Operations
            wrap_function_wrapper(
                "agno.workflow.workflow",
                "Workflow.run",
                workflow_run_wrap(
                    "workflow.run",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.workflow.workflow",
                "Workflow.run_workflow",
                workflow_run_wrap(
                    "workflow.run_workflow",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Async Workflow Operations (actual methods that exist)
            wrap_function_wrapper(
                "agno.workflow.workflow",
                "Workflow.arun",
                async_workflow_run_wrap(
                    "workflow.arun",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.workflow.workflow",
                "Workflow.arun_workflow",
                async_workflow_run_wrap(
                    "workflow.arun_workflow",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Team Operations
            wrap_function_wrapper(
                "agno.team.team",
                "Team.run",
                team_run_wrap(
                    "team.run",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Team Async Operations (actual methods that exist)
            wrap_function_wrapper(
                "agno.team.team",
                "Team.arun",
                async_team_run_wrap(
                    "team.arun",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Note: VectorDb class does not have async_search or async_upsert methods

            # Note: AgentKnowledge class does not have async_search method

            # Note: AgentKnowledge class does not have async_load or async_add_document_to_knowledge_base methods

            # NOTE: Removed duplicate tool execution instrumentation
            # The bridge spans from Model.run_function_call provide sufficient tracing
            # without creating redundant spans for FunctionCall.execute/aexecute

            # Reasoning Operations
            wrap_function_wrapper(
                "agno.agent.agent",
                "Agent.reason",
                reasoning_wrap(
                    "agent.reason",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.agent.agent",
                "Agent.areason",
                async_reasoning_wrap(
                    "agent.areason",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

        # Original Component Operations
        if detailed_tracing:
            wrap_function_wrapper(
                "agno.agent.agent",
                "Agent.add_tool",
                agent_add_tool_wrap(
                    "agent.add_tool",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Note: Agent class does not have async_add_tool method

            wrap_function_wrapper(
                "agno.agent.agent",
                "Agent.get_session_summary",
                session_memory_wrap(
                    "agent.get_session_summary",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            wrap_function_wrapper(
                "agno.agent.agent",
                "Agent.get_user_memories",
                session_memory_wrap(
                    "agent.get_user_memories",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # Note: Agent class does not have async_get_session_summary or async_get_user_memories methods

            # CRITICAL: Parallel execution operations for context preservation
            # Note: Parallel class doesn't have a run method, context is preserved via Model.run_function_calls

            wrap_function_wrapper(
                "agno.team.team",
                "Team._make_memories_and_summaries",
                parallel_execution_wrap(
                    "team._make_memories_and_summaries",
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    capture_message_content,
                    metrics,
                    disable_metrics,
                ),
            )

            # NOTE: Removed duplicate reasoning tool instrumentation
            # The bridge spans from Model.run_function_call provide sufficient tracing
            # without creating redundant spans for ReasoningTools.think/analyze

    def _patch_thread_pool_executor(self):
        """
        CRITICAL: Patch ThreadPoolExecutor.submit globally to preserve OpenTelemetry context.
        This fixes the broken span hierarchy caused by agno's extensive use of threading.
        """
        from concurrent.futures import ThreadPoolExecutor
        from opentelemetry import context as context_api

        # Store original submit method
        original_submit = ThreadPoolExecutor.submit

        def context_preserving_submit(self, func, *args, **kwargs):
            """Submit function while preserving OpenTelemetry context"""
            # Capture current context
            current_context = context_api.get_current()

            def context_wrapper(*wrapper_args, **wrapper_kwargs):
                # Restore context in the thread
                token = context_api.attach(current_context)
                try:
                    return func(*wrapper_args, **wrapper_kwargs)
                finally:
                    context_api.detach(token)

            return original_submit(self, context_wrapper, *args, **kwargs)

        # Apply the patch
        ThreadPoolExecutor.submit = context_preserving_submit

        # Store original for potential uninstrumentation
        self._original_thread_pool_submit = original_submit

    def _uninstrument(self, **kwargs):
        # Restore original ThreadPoolExecutor.submit if it was patched
        if hasattr(self, "_original_thread_pool_submit"):
            from concurrent.futures import ThreadPoolExecutor

            ThreadPoolExecutor.submit = self._original_thread_pool_submit



"""Initializer of Auto Instrumentation of Mistral Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.mistral.mistral import complete, stream, embed
from openlit.instrumentation.mistral.async_mistral import (
    async_complete,
    async_stream,
    async_embed,
)

_instruments = ("mistralai >= 1.0.0",)


class MistralInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Mistral client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default")
        environment = kwargs.get("environment", "default")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("mistralai")

        # sync chat completions
        wrap_function_wrapper(
            "mistralai.chat",
            "Chat.complete",
            complete(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync chat streaming
        wrap_function_wrapper(
            "mistralai.chat",
            "Chat.stream",
            stream(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync embeddings
        wrap_function_wrapper(
            "mistralai.embeddings",
            "Embeddings.create",
            embed(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # async chat completions
        wrap_function_wrapper(
            "mistralai.chat",
            "Chat.complete_async",
            async_complete(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # async chat streaming
        wrap_function_wrapper(
            "mistralai.chat",
            "Chat.stream_async",
            async_stream(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # async embeddings
        wrap_function_wrapper(
            "mistralai.embeddings",
            "Embeddings.create_async",
            async_embed(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

    def _uninstrument(self, **kwargs):
        pass



"""
Module for monitoring Azure AI Inference API calls.
"""

import logging
import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import (
    handle_exception,
    set_server_address_and_port,
)
from openlit.instrumentation.azure_ai_inference.utils import (
    process_chunk,
    process_chat_response,
    process_streaming_chat_response,
    process_embedding_response,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)


def complete(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    class TracedSyncStream:
        """
        Wrapper for streaming responses to collect telemetry.
        """

        def __init__(
            self,
            wrapped,
            span,
            span_name,
            kwargs,
            server_address,
            server_port,
            **args,
        ):
            self.__wrapped__ = wrapped
            self._span = span
            self._span_name = span_name
            self._llmresponse = ""
            self._response_id = ""
            self._response_model = ""
            self._finish_reason = ""
            self._response_service_tier = ""
            self._tools = None
            self._input_tokens = 0
            self._output_tokens = 0
            self._reasoning_tokens = 0

            self._args = args
            self._kwargs = kwargs
            self._start_time = time.time()
            self._end_time = None
            self._timestamps = []
            self._ttft = 0
            self._tbt = 0
            self._server_address = server_address
            self._server_port = server_port

        def __enter__(self):
            self.__wrapped__.__enter__()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)
            process_streaming_chat_response(
                self,
                pricing_info,
                environment,
                application_name,
                metrics,
                capture_message_content,
                disable_metrics,
                version,
            )

        def __iter__(self):
            return self

        def __next__(self):
            chunk = next(self.__wrapped__)
            process_chunk(self, chunk)
            return chunk

        def __getattr__(self, name):
            return getattr(self.__wrapped__, name)

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI function call.
        """

        streaming = kwargs.get("stream", False)
        server_address, server_port = set_server_address_and_port(
            instance, "models.github.ai", 443
        )
        request_model = kwargs.get("model", "gpt-4o")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        if streaming:
            awaited_wrapped = wrapped(*args, **kwargs)
            span = tracer.start_span(span_name, kind=SpanKind.CLIENT)

            return TracedSyncStream(
                awaited_wrapped, span, span_name, kwargs, server_address, server_port
            )

        else:
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                response = process_chat_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            return response

    return wrapper


def embed(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI embedding function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI embedding function call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "models.github.ai", 443
        )
        request_model = kwargs.get("model", "text-embedding-3-small")

        span_name = (
            f"{SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING} {request_model}"
        )

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = wrapped(*args, **kwargs)

            try:
                response = process_embedding_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper



"""
Utility functions for Agno instrumentation following OpenLIT patterns.
"""

import logging
import time
from openlit.__helpers import (
    common_span_attributes,
)
from openlit.semcov import SemanticConvention

# Initialize logger
logger = logging.getLogger(__name__)


# Simple utility functions following OpenLIT patterns


class _ScopeWrapper:
    """
    A scope wrapper class for common_span_attributes compatibility.

    This class wraps the necessary parameters required by the common_span_attributes
    function to maintain consistency with OpenLIT instrumentation patterns.
    """

    def __init__(self, span, instance, kwargs, args, start_time):
        """Initialize the scope wrapper with required parameters."""
        self._span = span
        self._instance = instance
        self._kwargs = kwargs
        self._args = args
        self._start_time = start_time
        self._end_time = time.time()

    def get_span(self):
        """Get the associated span."""
        return self._span

    def get_instance(self):
        """Get the instrumented instance."""
        return self._instance


def process_agent_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    operation_type=SemanticConvention.GEN_AI_OPERATION_TYPE_AGENT,
):
    """
    Process agent request and generate ALL telemetry attributes from semcov (CrewAI pattern)
    """

    # Calculate execution time
    execution_time = time.time() - start_time

    # Extract agent information
    agent_name = (
        getattr(instance, "name", None)
        or getattr(instance, "agent_id", None)
        or "default_agent"
    )
    request_model = None

    # Get model information
    if hasattr(instance, "model"):
        model = instance.model
        if hasattr(model, "id"):
            request_model = str(model.id)
        elif hasattr(model, "name"):
            request_model = str(model.name)

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes helper from __helpers
    common_span_attributes(
        scope,
        operation_type,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        request_model,
        request_model,  # response_model same as request
        environment,
        application_name,
        kwargs.get("stream", False),  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set agno-specific attributes using semantic conventions
    span.set_attribute(SemanticConvention.GEN_AI_AGENT_NAME, agent_name)

    # Set lifecycle phase based on operation
    if operation_type == SemanticConvention.GEN_AI_OPERATION_TYPE_AGENT:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_EXECUTE,
        )
    elif operation_type == SemanticConvention.GEN_AI_OPERATION_TYPE_MEMORY:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_EXECUTE,
        )
    elif operation_type == SemanticConvention.GEN_AI_OPERATION_TYPE_TOOLS:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
            SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_TOOL_EXECUTION,
        )

    # Add detailed agent information if available
    if hasattr(instance, "instructions") and instance.instructions:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_INSTRUCTIONS,
            str(instance.instructions)[:500],
        )

    if hasattr(instance, "agent_id") and instance.agent_id:
        span.set_attribute(SemanticConvention.GEN_AI_AGENT_ID, instance.agent_id)

    if hasattr(instance, "description") and instance.description:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_DESCRIPTION, str(instance.description)[:500]
        )

    if hasattr(instance, "introduction") and instance.introduction:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_INTRODUCTION,
            str(instance.introduction)[:500],
        )

    # Add request-specific attributes from args and kwargs
    if args and args[0]:
        if capture_message_content:
            span.set_attribute(
                SemanticConvention.GEN_AI_CONTENT_PROMPT, str(args[0])[:1000]
            )

    # User and session information
    if "user_id" in kwargs and kwargs["user_id"]:
        span.set_attribute(SemanticConvention.GEN_AI_REQUEST_USER, kwargs["user_id"])
    elif hasattr(instance, "user_id") and instance.user_id:
        span.set_attribute(SemanticConvention.GEN_AI_REQUEST_USER, instance.user_id)

    if "session_id" in kwargs and kwargs["session_id"]:
        span.set_attribute(SemanticConvention.GEN_AI_SESSION_ID, kwargs["session_id"])

    # Stream and reasoning attributes
    if "stream" in kwargs:
        span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_IS_STREAM, bool(kwargs["stream"])
        )

    if "show_full_reasoning" in kwargs:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_SHOW_REASONING,
            bool(kwargs["show_full_reasoning"]),
        )

    if "stream_intermediate_steps" in kwargs:
        span.set_attribute(
            SemanticConvention.GEN_AI_AGENT_STREAM_INTERMEDIATE_STEPS,
            bool(kwargs["stream_intermediate_steps"]),
        )

    # Set execution time using semantic convention
    span.set_attribute(SemanticConvention.GEN_AI_AGENT_EXECUTION_TIME, execution_time)

    # Process response with comprehensive output capture
    if response:
        if capture_message_content:
            if hasattr(response, "content") and response.content:
                content = str(response.content)[:2000]
                span.set_attribute(
                    SemanticConvention.GEN_AI_CONTENT_COMPLETION, content
                )
            elif hasattr(response, "message") and response.message:
                content = str(response.message)[:2000]
                span.set_attribute(
                    SemanticConvention.GEN_AI_CONTENT_COMPLETION, content
                )

        # Capture additional response metadata
        if hasattr(response, "metrics"):
            span.set_attribute(
                SemanticConvention.GEN_AI_AGENT_RESPONSE_TIME, execution_time
            )


def process_tool_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    tool_name,
):
    """
    Process tool execution request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Get model info from instance if available
    request_model = getattr(instance, "model", None)
    if request_model and hasattr(request_model, "id"):
        model_name = request_model.id
    else:
        model_name = "unknown"

    # Use common span attributes for tool operations
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        model_name,  # request_model
        model_name,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set tool-specific attributes using semcov
    span.set_attribute(SemanticConvention.GEN_AI_TOOL_NAME, tool_name)
    span.set_attribute(
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_TOOL_EXECUTION,
    )

    # Set tool-specific attributes that aren't in the process function
    if (
        hasattr(instance, "function")
        and hasattr(instance.function, "description")
        and instance.function.description
    ):
        span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_DESCRIPTION,
            str(instance.function.description)[:500],
        )

    if hasattr(instance, "arguments") and instance.arguments:
        span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_PARAMETERS, str(instance.arguments)[:1000]
        )

    # Set execution metrics
    span.set_attribute(
        SemanticConvention.GEN_AI_TOOL_EXECUTION_DURATION, execution_time
    )

    # Process result
    if response and hasattr(response, "status"):
        span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_EXECUTION_SUCCESS,
            response.status == "success",
        )
        if capture_message_content:
            if hasattr(response, "result") and response.result:
                span.set_attribute(
                    SemanticConvention.GEN_AI_TOOL_OUTPUT, str(response.result)[:1000]
                )
        if hasattr(response, "error") and response.error:
            span.set_attribute(
                SemanticConvention.GEN_AI_TOOL_ERROR, str(response.error)[:500]
            )


def process_memory_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    memory_operation,
):
    """
    Process memory operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes for memory operations
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_MEMORY,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        None,  # request_model
        None,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set memory-specific attributes using semcov
    span.set_attribute(SemanticConvention.GEN_AI_MEMORY_OPERATION, memory_operation)
    span.set_attribute(
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_EXECUTE,
    )

    # Set memory-specific attributes that aren't in the process function
    if hasattr(instance, "db") and instance.db:
        span.set_attribute("gen_ai.memory.db_type", type(instance.db).__name__)

    if hasattr(instance, "table_name"):
        span.set_attribute("gen_ai.memory.table_name", str(instance.table_name))

    # Add input data
    if args and capture_message_content:
        span.set_attribute(
            "gen_ai.memory.input", str(args[0])[:1000] if args[0] else ""
        )

    # Add metadata from kwargs
    if "user_id" in kwargs:
        span.set_attribute("gen_ai.memory.user_id", kwargs["user_id"])
    if "agent_id" in kwargs:
        span.set_attribute("gen_ai.memory.agent_id", kwargs["agent_id"])
    if "metadata" in kwargs:
        span.set_attribute("gen_ai.memory.metadata", str(kwargs["metadata"])[:500])

    # Set execution metrics
    span.set_attribute("gen_ai.memory.operation.duration", execution_time)

    # Process result
    if response:
        if hasattr(response, "id"):
            span.set_attribute("gen_ai.memory.result_id", str(response.id))
        span.set_attribute("gen_ai.memory.operation.success", True)
    else:
        span.set_attribute("gen_ai.memory.operation.success", False)


def process_reasoning_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    operation_type,
):
    """
    Process reasoning operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time
    agent_name = getattr(instance, "name", None) or "unknown_agent"

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Get model info from agent instance if available
    request_model = getattr(instance, "model", None)
    if request_model and hasattr(request_model, "id"):
        model_name = request_model.id
    else:
        model_name = "unknown"

    # Use common span attributes for reasoning operations (use agent task operation type)
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        model_name,  # request_model
        model_name,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set reasoning-specific attributes using semcov
    span.set_attribute(
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE,
        SemanticConvention.GEN_AI_AGENT_LIFECYCLE_PHASE_EXECUTE,
    )
    span.set_attribute(SemanticConvention.GEN_AI_AGENT_NAME, agent_name)

    # Add reasoning-specific attributes
    if hasattr(instance, "reasoning_min_steps"):
        span.set_attribute(
            SemanticConvention.GEN_AI_REASONING_MIN_STEPS, instance.reasoning_min_steps
        )

    if hasattr(instance, "reasoning_max_steps"):
        span.set_attribute(
            SemanticConvention.GEN_AI_REASONING_MAX_STEPS, instance.reasoning_max_steps
        )

    if hasattr(instance, "reasoning_model") and instance.reasoning_model:
        reasoning_model_name = getattr(
            instance.reasoning_model, "name", str(instance.reasoning_model)
        )
        span.set_attribute(
            SemanticConvention.GEN_AI_REASONING_MODEL, reasoning_model_name
        )

    # Set execution time
    span.set_attribute(
        SemanticConvention.GEN_AI_REASONING_EXECUTION_DURATION, execution_time
    )


def process_vectordb_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    operation_type="vectordb_search",
):
    """
    Process vectordb operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes for vectordb operations
    common_span_attributes(
        scope,
        f"gen_ai.operation.{operation_type}",
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        None,  # request_model
        None,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set vectordb-specific attributes using semcov
    if hasattr(instance, "name"):
        span.set_attribute(SemanticConvention.GEN_AI_VECTORDB_NAME, str(instance.name))

    if hasattr(instance, "dimensions"):
        span.set_attribute(
            SemanticConvention.GEN_AI_VECTORDB_DIMENSIONS, instance.dimensions
        )

    # Search-specific attributes
    if operation_type == "vectordb_search" and args:
        query = args[0] if args else None
        if query:
            if isinstance(query, str):
                span.set_attribute(
                    SemanticConvention.GEN_AI_VECTORDB_SEARCH_QUERY, query[:500]
                )
            elif hasattr(query, "__len__"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_VECTORDB_SEARCH_VECTOR_SIZE, len(query)
                )

        limit = kwargs.get("limit", args[1] if len(args) > 1 else None)
        if limit:
            span.set_attribute(SemanticConvention.GEN_AI_VECTORDB_SEARCH_LIMIT, limit)

    # Upsert-specific attributes
    if operation_type == "vectordb_upsert" and args:
        documents = args[0] if args else None
        if documents and hasattr(documents, "__len__"):
            span.set_attribute(
                SemanticConvention.GEN_AI_VECTORDB_UPSERT_DOCUMENT_COUNT, len(documents)
            )

    # Set execution metrics
    span.set_attribute(
        SemanticConvention.GEN_AI_VECTORDB_OPERATION_DURATION, execution_time
    )

    # Process result
    if response:
        span.set_attribute(SemanticConvention.GEN_AI_VECTORDB_OPERATION_SUCCESS, True)
        if operation_type == "vectordb_search" and hasattr(response, "__len__"):
            span.set_attribute(
                SemanticConvention.GEN_AI_VECTORDB_SEARCH_RESULTS_COUNT, len(response)
            )
            if response and hasattr(response[0], "score"):
                scores = [r.score for r in response[:5]]  # Top 5 scores
                span.set_attribute(
                    SemanticConvention.GEN_AI_VECTORDB_SEARCH_TOP_SCORES, scores
                )
    else:
        span.set_attribute(SemanticConvention.GEN_AI_VECTORDB_OPERATION_SUCCESS, False)


def process_knowledge_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    operation_type="knowledge_search",
):
    """
    Process knowledge operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes for knowledge operations
    common_span_attributes(
        scope,
        f"gen_ai.operation.{operation_type}",
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        None,  # request_model
        None,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Search-specific attributes
    if operation_type == "knowledge_search":
        query = args[0] if args else None
        if query and capture_message_content:
            span.set_attribute(
                SemanticConvention.GEN_AI_KNOWLEDGE_SEARCH_QUERY, query[:500]
            )

        limit = kwargs.get("limit", args[1] if len(args) > 1 else 5)
        span.set_attribute(SemanticConvention.GEN_AI_KNOWLEDGE_SEARCH_LIMIT, limit)

    # Add-specific attributes
    if operation_type == "knowledge_add" and args:
        documents = args[0] if args else None
        if documents:
            if hasattr(documents, "__len__"):
                span.set_attribute(
                    SemanticConvention.GEN_AI_KNOWLEDGE_ADD_DOCUMENT_COUNT,
                    len(documents),
                )
            else:
                span.set_attribute(
                    SemanticConvention.GEN_AI_KNOWLEDGE_ADD_CONTENT_LENGTH,
                    len(str(documents)),
                )

    # Set execution metrics
    span.set_attribute(
        SemanticConvention.GEN_AI_KNOWLEDGE_OPERATION_DURATION, execution_time
    )

    # Process result
    if response:
        span.set_attribute(SemanticConvention.GEN_AI_KNOWLEDGE_OPERATION_SUCCESS, True)
        if operation_type == "knowledge_search" and hasattr(response, "__len__"):
            span.set_attribute(
                SemanticConvention.GEN_AI_KNOWLEDGE_SEARCH_RESULTS_COUNT, len(response)
            )
    else:
        span.set_attribute(SemanticConvention.GEN_AI_KNOWLEDGE_OPERATION_SUCCESS, False)


def process_workflow_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Process workflow operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time
    workflow_name = getattr(instance, "name", "unknown_workflow")

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes for workflow operations
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_WORKFLOW,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        None,  # request_model
        None,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set workflow-specific attributes using semcov
    span.set_attribute(SemanticConvention.GEN_AI_WORKFLOW_NAME, workflow_name)

    if hasattr(instance, "agents") and instance.agents:
        span.set_attribute(
            SemanticConvention.GEN_AI_WORKFLOW_AGENT_COUNT, len(instance.agents)
        )

    if hasattr(instance, "description") and instance.description:
        span.set_attribute(
            SemanticConvention.GEN_AI_WORKFLOW_DESCRIPTION,
            str(instance.description)[:300],
        )

    # Set execution metrics
    span.set_attribute(
        SemanticConvention.GEN_AI_WORKFLOW_EXECUTION_DURATION, execution_time
    )
    span.set_attribute(
        SemanticConvention.GEN_AI_WORKFLOW_OPERATION_SUCCESS, response is not None
    )


def process_team_request(
    span,
    instance,
    args,
    kwargs,
    response,
    start_time,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Process team operation request and generate ALL telemetry attributes from semcov
    """

    execution_time = time.time() - start_time
    team_name = getattr(instance, "name", "unknown_team")

    # Create scope object for common_span_attributes compatibility
    scope = _ScopeWrapper(span, instance, kwargs, args, start_time)

    # Use common span attributes for team operations
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_TEAM,
        SemanticConvention.GEN_AI_SYSTEM_AGNO,
        "localhost",  # server_address
        80,  # server_port
        None,  # request_model
        None,  # response_model
        environment,
        application_name,
        False,  # is_stream
        0,  # tbt
        0,  # ttft
        version,
    )

    # Set team-specific attributes using semcov
    span.set_attribute(SemanticConvention.GEN_AI_TEAM_NAME, team_name)

    if hasattr(instance, "agents") and instance.agents:
        agent_names = [getattr(agent, "name", "unknown") for agent in instance.agents]
        span.set_attribute(
            SemanticConvention.GEN_AI_TEAM_AGENTS, agent_names[:10]
        )  # Limit to 10
        span.set_attribute(
            SemanticConvention.GEN_AI_TEAM_AGENT_COUNT, len(instance.agents)
        )

    # Set execution metrics
    span.set_attribute(
        SemanticConvention.GEN_AI_TEAM_EXECUTION_DURATION, execution_time
    )
    span.set_attribute(
        SemanticConvention.GEN_AI_TEAM_OPERATION_SUCCESS, response is not None
    )



# pylint: disable=duplicate-code, line-too-long, too-few-public-methods, too-many-instance-attributes
"""
Module for finding Hallucination in text.
"""

from typing import Optional, List, Dict
from openlit.evals.utils import (
    setup_provider,
    JsonOutput,
    format_prompt,
    llm_response,
    parse_llm_response,
    eval_metrics,
    eval_metric_attributes,
)


# pylint: disable=unused-argument
def get_system_prompt(
    custom_categories: Optional[Dict[str, str]] = None,
    threshold_score: Optional[float] = 0.5,
) -> str:
    """
    Returns the system prompt used for LLM analysis, including custom categories if provided.

    Args:
        prompt (Optional[str]): The prompt provided by the user.
        contexts (Optional[List[str]]): A list of context sentences relevant to the task.
        text (Optional[str]): The text to analyze.
        custom_categories (Optional[Dict[str, str]]): Additional categories for hallucination.
        threshold_score (Optional[float]): The threshold score to determine the verdict.

    Returns:
        str: The system prompt string.
    """
    base_prompt = """
    Output Requirement: JSON Format Only

    Your task is to find any instances of Hallucination in text compared to the provided contexts and the optional prompt. Generate a JSON object with the following fields: 'score', 'evaluation', 'classification', 'explanation', and 'verdict'. Use the contexts to strictly detect hallucination in the text.

    - 'score': A float from 0 to 1. Closer to 1 indicates a higher level of hallucination.
    - 'evaluation': Should always be "hallucination".
    - 'classification': Type of Hallucination from one of the hallucination categories.
    - 'explanation': Provide a very short sentence explanation for the selected category of Hallucination.
    - 'verdict': 'yes' if score > {threshold_score}, otherwise 'no'.

    Hallucination Categories:
    - factual_inaccuracy: Incorrect facts, e.g., Context: ["Paris is the capital of France."]; Text: "Lyon is the capital."
    - nonsensical_response: Irrelevant info, e.g., Context: ["Discussing music trends."]; Text: "Golf uses clubs on grass."
    - gibberish: Nonsensical text, e.g., Context: ["Discuss advanced algorithms."]; Text: "asdas asdhasudqoiwjopakcea."
    - contradiction: Conflicting info, e.g., Context: ["Einstein was born in 1879."]; Text: "Einstein was born in 1875 and 1879."
    """

    if custom_categories:
        custom_categories_str = "\n".join(
            [f"- {key}: {value}" for key, value in custom_categories.items()]
        )
        base_prompt += (
            f"\n    Additional Hallucination Categories:\n{custom_categories_str}"
        )

    base_prompt += """

    For each context provided, verify if the 'text' aligns with it, and score the text based on the level of alignment or contradiction:

    - If a context is contradicted, increase the hallucination score. Provide specific contradictions in the explanation.
    - Forgive minor omissions but classify as contradiction only when there is a clear factual inconsistency or significant contextual mismatch.

    **
    IMPORTANT: Return JSON format only.

    Example contexts: ["Einstein won the Nobel Prize for his discovery of the photoelectric effect.", "Einstein won the Nobel Prize in 1968."]
    Text: "Einstein won the Nobel Prize in 1969 for his discovery of the photoelectric effect."

    Example JSON:
    {
        "score": 0.8,
        "evaluation": "hallucination",
        "classification": "factual_inaccuracy",
        "explanation": "The output has incorrect facts",
        "verdict": "yes"
    }

    **
    prompt (Optional. Only take into context if provided.):
    {{prompt}}

    Contexts:
    {{context}}

    Text:
    {{text}}

    JSON:

    If no hallucination is detected, return:
    {
        "score": 0,
        "evaluation": "hallucination",
        "classification": "none",
        "explanation": "All text is consistent with context.",
        "verdict": "no"
    }
    """

    return base_prompt


class Hallucination:
    """
    A class to detect hallucinations in AI responses using LLM or custom categories.

    Attributes:
        provider (Optional[str]): The name of the LLM provider.
        api_key (Optional[str]): The API key for authenticating with the LLM.
        model (Optional[str]): The name of the model to use in the LLM.
        base_url (Optional[str]): The base URL for the LLM API.
        custom_categories (Optional[Dict[str, str]]): Additional categories for prompt injections.
    """

    def __init__(
        self,
        provider: Optional[str] = "openai",
        api_key: Optional[str] = None,
        model: Optional[str] = None,
        base_url: Optional[str] = None,
        custom_categories: Optional[Dict[str, str]] = None,
        collect_metrics: Optional[bool] = False,
        threshold_score: Optional[float] = 0.5,
    ):
        """
        Initializes the Hallucination detector with specified LLM settings, custom rules, and categories.

        Args:
            provider (Optional[str]): The name of the LLM provider.
            api_key (Optional[str]): The API key for authenticating with the LLM.
            model (Optional[str]): The name of the model to use in the LLM.
            base_url (Optional[str]): The base URL for the LLM API.
            threshold_score (float): User-defined threshold to determine the verdict.

        Raises:
            ValueError: If provider is not specified.
        """

        self.provider = provider
        if self.provider is None:
            raise ValueError(
                "An LLM provider must be specified for Hallucination detection."
            )
        self.api_key, self.model, self.base_url = setup_provider(
            provider, api_key, model, base_url
        )
        self.collect_metrics = collect_metrics
        self.custom_categories = custom_categories
        self.threshold_score = threshold_score
        self.system_prompt = get_system_prompt(
            self.custom_categories, self.threshold_score
        )

    def measure(
        self,
        prompt: Optional[str] = "",
        contexts: Optional[List[str]] = None,
        text: Optional[str] = None,
    ) -> JsonOutput:
        """
        Detects hallucinations in AI output using LLM or custom rules.

        Args:
            prompt (Optional[str]): The prompt provided by the user.
            contexts (Optional[List[str]]): A list of context sentences relevant to the task.
            text (Optional[str]): The text to analyze.

        Returns:
            JsonOutput: The result containing score, evaluation, classification, explanation, and verdict of hallucination detection.
        """

        llm_prompt = format_prompt(self.system_prompt, prompt, contexts, text)
        response = llm_response(self.provider, llm_prompt, self.model, self.base_url)
        llm_result = parse_llm_response(response)
        result_verdict = "yes" if llm_result.score > self.threshold_score else "no"
        result = JsonOutput(
            score=llm_result.score,
            evaluation=llm_result.evaluation,
            classification=llm_result.classification,
            explanation=llm_result.explanation,
            verdict=result_verdict,
        )

        if self.collect_metrics:
            eval_counter = eval_metrics()
            attributes = eval_metric_attributes(
                result_verdict,
                result.score,
                result.evaluation,
                result.classification,
                result.explanation,
            )
            eval_counter.add(1, attributes)

        return result



"""Initializer of Auto Instrumentation of AG2 Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.ag2.ag2 import (
    conversable_agent,
    agent_run,
    agent_generate_reply,
    agent_receive,
    agent_send,
    groupchat_manager_run_chat,
    groupchat_select_speaker,
)
from openlit.instrumentation.ag2.async_ag2 import (
    async_conversable_agent,
    async_agent_run,
    async_agent_generate_reply,
    async_agent_receive,
    async_agent_send,
)

_instruments = ("ag2 >= 0.3.2",)


class AG2Instrumentor(BaseInstrumentor):
    """
    An instrumentor for AG2 client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        version = importlib.metadata.version("ag2")
        environment = kwargs.get("environment", "default")
        application_name = kwargs.get("application_name", "default")
        tracer = kwargs.get("tracer")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        metrics = kwargs.get("metrics_dict")
        disable_metrics = kwargs.get("disable_metrics")

        # sync conversable agent
        wrap_function_wrapper(
            "autogen.agentchat.conversable_agent",
            "ConversableAgent.__init__",
            conversable_agent(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync agent run
        wrap_function_wrapper(
            "autogen.agentchat.conversable_agent",
            "ConversableAgent.run",
            agent_run(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync agent generate_reply
        wrap_function_wrapper(
            "autogen.agentchat.conversable_agent",
            "ConversableAgent.generate_reply",
            agent_generate_reply(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync agent receive
        wrap_function_wrapper(
            "autogen.agentchat.conversable_agent",
            "ConversableAgent.receive",
            agent_receive(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync agent send
        wrap_function_wrapper(
            "autogen.agentchat.conversable_agent",
            "ConversableAgent.send",
            agent_send(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync groupchat manager run_chat
        wrap_function_wrapper(
            "autogen.agentchat.groupchat",
            "GroupChatManager.run_chat",
            groupchat_manager_run_chat(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

        # sync groupchat select_speaker
        wrap_function_wrapper(
            "autogen.agentchat.groupchat",
            "GroupChat.select_speaker",
            groupchat_select_speaker(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics,
            ),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
ChromaDB OpenTelemetry instrumentation utility functions
"""

import time
from urllib.parse import urlparse
from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    common_db_span_attributes,
    record_db_metrics,
)
from openlit.semcov import SemanticConvention

# Operation mapping for simple span naming
DB_OPERATION_MAP = {
    "chroma.create_collection": SemanticConvention.DB_OPERATION_CREATE_COLLECTION,
    "chroma.add": SemanticConvention.DB_OPERATION_INSERT,
    "chroma.get": SemanticConvention.DB_OPERATION_GET,
    "chroma.peek": SemanticConvention.DB_OPERATION_PEEK,
    "chroma.query": SemanticConvention.DB_OPERATION_GET,
    "chroma.update": SemanticConvention.DB_OPERATION_UPDATE,
    "chroma.upsert": SemanticConvention.DB_OPERATION_UPSERT,
    "chroma.delete": SemanticConvention.DB_OPERATION_DELETE,
}


def object_count(obj):
    """
    Counts length of object if it exists, else returns 0.
    """
    return len(obj) if obj else 0


def set_server_address_and_port(instance):
    """
    Extracts server address and port from ChromaDB client instance.

    Args:
        instance: ChromaDB client instance

    Returns:
        tuple: (server_address, server_port)
    """
    server_address = "localhost"
    server_port = 8000

    # Try getting base_url from multiple potential attributes
    base_client = getattr(instance, "_client", None)
    base_url = getattr(base_client, "base_url", None)

    if not base_url:
        # Attempt to get endpoint from instance._config.endpoint
        config = getattr(instance, "_config", None)
        base_url = getattr(config, "endpoint", None)

    if not base_url:
        # Attempt to get server_url from instance.sdk_configuration.server_url
        config = getattr(instance, "sdk_configuration", None)
        base_url = getattr(config, "server_url", None)

    if base_url:
        if isinstance(base_url, str):
            # Check if it a full URL or just a hostname
            if base_url.startswith(("http://", "https://")):
                try:
                    url = urlparse(base_url)
                    if url.hostname:
                        server_address = url.hostname
                    if url.port:
                        server_port = url.port
                except Exception:
                    pass
            else:
                # Just a hostname
                server_address = base_url

    return server_address, server_port


def common_vectordb_logic(
    scope,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    instance=None,
    endpoint=None,
):
    """
    Process vector database request and generate telemetry.
    """

    scope._end_time = time.time()

    # Set common database span attributes using helper
    common_db_span_attributes(
        scope,
        SemanticConvention.DB_SYSTEM_CHROMA,
        scope._server_address,
        scope._server_port,
        environment,
        application_name,
        version,
    )

    # Set DB operation specific attributes
    scope._span.set_attribute(SemanticConvention.DB_OPERATION_NAME, scope._db_operation)
    scope._span.set_attribute(
        SemanticConvention.DB_CLIENT_OPERATION_DURATION,
        scope._end_time - scope._start_time,
    )

    # Set collection name from instance
    if hasattr(instance, "name"):
        scope._span.set_attribute(SemanticConvention.DB_COLLECTION_NAME, instance.name)

    # Set Create Collection operation specific attributes
    if scope._db_operation == SemanticConvention.DB_OPERATION_CREATE_COLLECTION:
        # Standard database attributes
        collection_name = scope._kwargs.get("name") or (
            scope._args[0] if scope._args else "unknown"
        )
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )

        # Vector database specific attributes (extensions)
        metadata = scope._kwargs.get("metadata", {})
        if metadata:
            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_METADATA, str(metadata)
            )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_INSERT:
        collection_name = getattr(instance, "name", "unknown")
        query = scope._kwargs.get("ids", [])

        # Standard database attributes
        scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_VECTOR_COUNT, object_count(query)
        )

        # Vector database specific attributes (extensions)
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} "
            f"ids={query} "
            f"documents={scope._kwargs.get('documents', [])}",
        )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_GET:
        collection_name = getattr(instance, "name", "unknown")

        # Handle different GET operations based on endpoint
        if endpoint == "chroma.get":
            # Collection.get() - retrieve documents by IDs
            query = scope._kwargs.get("ids", [])

            # Standard database attributes
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))
            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(query)
            )

            # Vector database specific attributes (extensions)
            scope._span.set_attribute(
                SemanticConvention.DB_FILTER, str(scope._kwargs.get("where", ""))
            )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"ids={query} "
                f"limit={scope._kwargs.get('limit', 'None')} "
                f"offset={scope._kwargs.get('offset', 'None')}",
            )

        elif endpoint == "chroma.query":
            query_texts = scope._kwargs.get("query_texts", [])
            query_embeddings = scope._kwargs.get("query_embeddings", [])

            # Create comprehensive query text (can be either embeddings or texts)
            if query_texts:
                query_content = f"texts={query_texts}"
            elif query_embeddings:
                query_content = f"embeddings={len(query_embeddings) if query_embeddings else 0} vectors"
            else:
                query_content = "no query provided"

            # Standard database attributes
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, query_content)
            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )

            # Vector database specific attributes (extensions)
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_QUERY_TOP_K,
                scope._kwargs.get("n_results", 10),
            )
            scope._span.set_attribute(
                SemanticConvention.DB_FILTER, str(scope._kwargs.get("where", ""))
            )

            # Extract response metrics if available
            if scope._response:
                # Get number of results returned
                if hasattr(scope._response, "get") and scope._response.get("ids"):
                    returned_rows = (
                        object_count(scope._response["ids"][0])
                        if scope._response["ids"]
                        else 0
                    )
                    scope._span.set_attribute(
                        SemanticConvention.DB_RESPONSE_RETURNED_ROWS, returned_rows
                    )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"n_results={scope._kwargs.get('n_results', 10)} "
                f"{query_content} "
                f"filter={scope._kwargs.get('where', 'None')}",
            )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_UPDATE:
        collection_name = getattr(instance, "name", "unknown")
        query = scope._kwargs.get("ids", [])

        # Standard database attributes
        scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_VECTOR_COUNT, object_count(query)
        )

        # Vector database specific attributes (extensions)
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} "
            f"ids={query} "
            f"embeddings={scope._kwargs.get('embeddings', 'None')} "
            f"metadatas={scope._kwargs.get('metadatas', 'None')} "
            f"documents={scope._kwargs.get('documents', 'None')}",
        )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_UPSERT:
        collection_name = getattr(instance, "name", "unknown")
        query = scope._kwargs.get("ids", [])

        # Standard database attributes
        scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_VECTOR_COUNT, object_count(query)
        )

        # Vector database specific attributes (extensions)
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} "
            f"ids={query} "
            f"embeddings={scope._kwargs.get('embeddings', 'None')} "
            f"metadatas={scope._kwargs.get('metadatas', 'None')} "
            f"documents={scope._kwargs.get('documents', 'None')}",
        )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_DELETE:
        collection_name = getattr(instance, "name", "unknown")
        query = scope._kwargs.get("ids", [])

        # Standard database attributes
        scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_VECTOR_COUNT, object_count(query)
        )

        # Vector database specific attributes (extensions)
        scope._span.set_attribute(
            SemanticConvention.DB_FILTER, str(scope._kwargs.get("where", ""))
        )
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} "
            f"ids={query} "
            f"filter={scope._kwargs.get('where', 'None')} "
            f"delete_all={scope._kwargs.get('delete_all', False)}",
        )

    elif scope._db_operation == SemanticConvention.DB_OPERATION_PEEK:
        collection_name = getattr(instance, "name", "unknown")
        query = f"PEEK limit={scope._kwargs.get('limit', '')}"

        # Standard database attributes
        scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, query)
        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_VECTOR_COUNT, scope._kwargs.get("limit", "")
        )

        # Vector database specific attributes (extensions)
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} "
            f"limit={scope._kwargs.get('limit', 'None')}",
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics using helper
    if not disable_metrics:
        record_db_metrics(
            metrics,
            SemanticConvention.DB_SYSTEM_CHROMA,
            scope._server_address,
            scope._server_port,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
        )


def process_vectordb_response(
    response,
    db_operation,
    server_address,
    server_port,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    instance=None,
    args=None,
    endpoint=None,
    **kwargs,
):
    """
    Process vector database response and generate telemetry following OpenTelemetry conventions.
    """

    scope = type("GenericScope", (), {})()

    scope._start_time = start_time
    scope._span = span
    scope._kwargs = kwargs
    scope._args = args or []
    scope._db_operation = db_operation
    scope._response = response
    scope._server_address = server_address
    scope._server_port = server_port

    common_vectordb_logic(
        scope,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        instance,
        endpoint,
    )

    return response



"""
vLLM OpenTelemetry instrumentation utility functions
"""

import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    general_tokens,
    get_chat_model_cost,
    common_span_attributes,
    record_completion_metrics,
)
from openlit.semcov import SemanticConvention


def get_inference_config(args, kwargs):
    """
    Safely extract inference configuration from args or kwargs.
    """

    if "sampling_params" in kwargs:
        return kwargs["sampling_params"]
    if len(args) > 1:
        return args[1]
    return None


def format_content(prompts):
    """
    Process a list of prompts to extract content.
    """

    if isinstance(prompts, str):
        return prompts
    elif isinstance(prompts, list):
        return "\n".join(str(prompt) for prompt in prompts)
    else:
        return str(prompts)


def common_chat_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process chat request and generate Telemetry
    """

    request_model = scope._request_model

    # Extract prompts and completions from vLLM response
    input_tokens = 0
    output_tokens = 0
    prompt = ""
    completion = ""

    for output in scope._response:
        prompt += output.prompt + "\n"
        if output.outputs and len(output.outputs) > 0:
            completion += output.outputs[0].text + "\n"
        input_tokens += general_tokens(output.prompt)
        output_tokens += general_tokens(output.outputs[0].text)

    cost = get_chat_model_cost(request_model, pricing_info, input_tokens, output_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_VLLM,
        scope._server_address,
        scope._server_port,
        request_model,
        request_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Request parameters
    inference_config = get_inference_config(scope._args, scope._kwargs)
    if inference_config:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
            getattr(inference_config, "max_tokens", -1),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES,
            getattr(inference_config, "stop_sequences", []),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
            getattr(inference_config, "temperature", 1.0),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_TOP_P,
            getattr(inference_config, "top_p", 1.0),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_TOP_K,
            getattr(inference_config, "top_k", -1),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY,
            getattr(inference_config, "presence_penalty", 0.0),
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY,
            getattr(inference_config, "frequency_penalty", 0.0),
        )

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "text")

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, input_tokens + output_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, completion
        )

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: completion,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_VLLM,
            scope._server_address,
            scope._server_port,
            request_model,
            request_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            input_tokens,
            output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def process_chat_response(
    instance,
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    args,
    kwargs,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
):
    """
    Process chat request and generate Telemetry
    """

    # Create scope object
    scope = type("GenericScope", (), {})()

    scope._response = response
    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address = server_address
    scope._server_port = server_port
    scope._request_model = request_model
    scope._timestamps = []
    scope._args = args
    scope._kwargs = kwargs

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response



"""
PremAI OpenTelemetry instrumentation utility functions
"""

import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    response_as_dict,
    calculate_ttft,
    calculate_tbt,
    get_chat_model_cost,
    get_embed_model_cost,
    general_tokens,
    common_span_attributes,
    record_completion_metrics,
    record_embedding_metrics,
)
from openlit.semcov import SemanticConvention


def format_content(messages):
    """
    Process a list of messages to extract content.
    """

    formatted_messages = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if isinstance(content, list):
            content_str = ", ".join(
                f"{item['type']}: {item['text'] if 'text' in item else item['image_url']}"
                if "type" in item
                else f"text: {item['text']}"
                for item in content
            )
            formatted_messages.append(f"{role}: {content_str}")
        else:
            formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)


def process_chunk(scope, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        # Calculate time to first chunk
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)
    # Collect message IDs and aggregated response from events
    if chunked.choices:
        first_choice = chunked.get("choices")[0]

        if first_choice.get("delta").get("content"):
            scope._llmresponse += first_choice.get("delta").get("content")

    if chunked.get("choices")[0].get("finish_reason"):
        scope._finish_reason = chunked.get("choices")[0].get("finish_reason")
        scope._response_id = chunked.get("id")
        scope._response_model = chunked.get("model")
        scope._end_time = time.time()


def common_chat_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    is_stream,
):
    """
    Process chat request and generate Telemetry
    """

    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    prompt = format_content(scope._kwargs.get("messages", ""))
    request_model = scope._kwargs.get("model", "llama3.2-3b")

    # Calculate tokens using input prompt and aggregated response
    if is_stream:
        input_tokens = general_tokens(prompt)
        output_tokens = general_tokens(scope._llmresponse)
    else:
        input_tokens = scope._input_tokens
        output_tokens = scope._output_tokens

    cost = get_chat_model_cost(request_model, pricing_info, input_tokens, output_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
        SemanticConvention.GEN_AI_SYSTEM_PREMAI,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        is_stream,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Span Attributes for Response parameters
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason]
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_SEED, scope._kwargs.get("seed", "")
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY,
        scope._kwargs.get("frequency_penalty", 0.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS,
        scope._kwargs.get("max_tokens", -1),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY,
        scope._kwargs.get("presence_penalty", 0.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get("stop", [])
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
        scope._kwargs.get("temperature", 1.0),
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get("top_p", 1.0)
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_USER, scope._kwargs.get("user", "")
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_OUTPUT_TYPE,
        "text" if isinstance(scope._llmresponse, str) else "json",
    )

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, input_tokens + output_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Tools
    if scope._tools:
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_NAME, scope._tools.get("function", "")
        ).get("name", "")
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_CALL_ID, str(scope._tools.get("id", ""))
        )
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_TOOL_ARGS,
            str(scope._tools.get("function", "").get("arguments", "")),
        )

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(
            SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse
        )

        # To be removed one the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_completion_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            SemanticConvention.GEN_AI_SYSTEM_PREMAI,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            input_tokens,
            output_tokens,
            cost,
            scope._tbt,
            scope._ttft,
        )


def common_embedding_logic(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
):
    """
    Process embedding request and generate Telemetry
    """

    request_model = scope._kwargs.get("model", "text-embedding-ada-002")

    cost = get_embed_model_cost(request_model, pricing_info, scope._input_tokens)

    # Common Span Attributes
    common_span_attributes(
        scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING,
        SemanticConvention.GEN_AI_SYSTEM_PREMAI,
        scope._server_address,
        scope._server_port,
        request_model,
        scope._response_model,
        environment,
        application_name,
        False,
        scope._tbt,
        scope._ttft,
        version,
    )

    # Embedding-specific span attributes
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_ENCODING_FORMATS,
        [scope._kwargs.get("encoding_format", "float")],
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_REQUEST_USER, scope._kwargs.get("user", "")
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens
    )
    scope._span.set_attribute(
        SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens
    )
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: str(
                    scope._kwargs.get("input", "")
                ),
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_embedding_metrics(
            metrics,
            SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING,
            SemanticConvention.GEN_AI_SYSTEM_PREMAI,
            scope._server_address,
            scope._server_port,
            request_model,
            scope._response_model,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
            scope._input_tokens,
            cost,
        )


def process_streaming_chat_response(
    scope,
    pricing_info,
    environment,
    application_name,
    metrics,
    capture_message_content=False,
    disable_metrics=False,
    version="",
):
    """
    Process chat request and generate Telemetry
    """
    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=True,
    )


def process_chat_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process chat request and generate Telemetry
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = str(
        response_dict.get("choices")[0].get("message").get("content")
    )
    scope._response_id = response_dict.get("additional_properties", {}).get("id")
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage").get("prompt_tokens")
    scope._output_tokens = response_dict.get("usage").get("completion_tokens")
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._finish_reason = str(response_dict.get("choices")[0].get("finish_reason"))

    if scope._kwargs.get("tools"):
        scope._tools = response_dict.get("choices")[0].get("message").get("tool_calls")
    else:
        scope._tools = None

    common_chat_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        is_stream=False,
    )

    return response


def process_embedding_response(
    response,
    request_model,
    pricing_info,
    server_port,
    server_address,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content=False,
    disable_metrics=False,
    version="1.0.0",
    **kwargs,
):
    """
    Process embedding request and generate Telemetry
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage").get("prompt_tokens")
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    common_embedding_logic(
        scope,
        pricing_info,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
    )

    return response



"""
Module for monitoring HF Transformers API calls.
"""

import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import set_server_address_and_port
from openlit.instrumentation.transformers.utils import process_chat_response
from openlit.semcov import SemanticConvention


def pipeline_wrapper(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI function call.
        """

        server_address, server_port = set_server_address_and_port(
            instance, "127.0.0.1", 80
        )
        request_model = instance.model.config.name_or_path

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = wrapped(*args, **kwargs)

            response = process_chat_response(
                instance=instance,
                response=response,
                request_model=request_model,
                pricing_info=pricing_info,
                server_port=server_port,
                server_address=server_address,
                environment=environment,
                application_name=application_name,
                metrics=metrics,
                start_time=start_time,
                span=span,
                args=args,
                kwargs=kwargs,
                capture_message_content=capture_message_content,
                disable_metrics=disable_metrics,
                version=version,
            )

        return response

    return wrapper



"""
Module for monitoring LiteLLM API calls (async version).
"""

import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import handle_exception, set_server_address_and_port
from openlit.instrumentation.litellm.utils import (
    process_chunk,
    process_streaming_chat_response,
    process_chat_response,
    process_embedding_response,
)
from openlit.semcov import SemanticConvention


def acompletion(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    class TracedAsyncStream:
        """
        Wrapper for async streaming responses to collect telemetry.
        """

        def __init__(
            self,
            wrapped,
            span,
            span_name,
            kwargs,
            server_address,
            server_port,
            **args,
        ):
            self.__wrapped__ = wrapped
            self._span = span
            self._span_name = span_name
            self._llmresponse = ""
            self._response_id = ""
            self._response_model = ""
            self._finish_reason = ""
            self._response_service_tier = ""
            self._tools = None
            self._input_tokens = 0
            self._output_tokens = 0
            self._args = args
            self._kwargs = kwargs
            self._start_time = time.time()
            self._end_time = None
            self._timestamps = []
            self._ttft = 0
            self._tbt = 0
            self._server_address = server_address
            self._server_port = server_port

        async def __aenter__(self):
            await self.__wrapped__.__aenter__()
            return self

        async def __aexit__(self, exc_type, exc_value, traceback):
            await self.__wrapped__.__aexit__(exc_type, exc_value, traceback)

        def __aiter__(self):
            return self

        async def __getattr__(self, name):
            """Delegate attribute access to the wrapped object."""
            return getattr(await self.__wrapped__, name)

        async def __anext__(self):
            try:
                chunk = await self.__wrapped__.__anext__()
                process_chunk(self, chunk)
                return chunk
            except StopAsyncIteration:
                try:
                    with tracer.start_as_current_span(
                        self._span_name, kind=SpanKind.CLIENT
                    ) as self._span:
                        process_streaming_chat_response(
                            self,
                            pricing_info=pricing_info,
                            environment=environment,
                            application_name=application_name,
                            metrics=metrics,
                            capture_message_content=capture_message_content,
                            disable_metrics=disable_metrics,
                            version=version,
                        )

                except Exception as e:
                    handle_exception(self._span, e)

                raise

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI function call.
        """
        # Check if streaming is enabled for the API call
        streaming = kwargs.get("stream", False)
        server_address, server_port = set_server_address_and_port(
            instance, "NOT_FOUND", "NOT_FOUND"
        )
        request_model = kwargs.get("model", "openai/gpt-4o")

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        if streaming:
            # Special handling for streaming response
            awaited_wrapped = await wrapped(*args, **kwargs)
            span = tracer.start_span(span_name, kind=SpanKind.CLIENT)
            return TracedAsyncStream(
                awaited_wrapped, span, span_name, kwargs, server_address, server_port
            )
        else:
            # Handling for non-streaming responses
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = await wrapped(*args, **kwargs)

                try:
                    response = process_chat_response(
                        response=response,
                        request_model=request_model,
                        pricing_info=pricing_info,
                        server_port=server_port,
                        server_address=server_address,
                        environment=environment,
                        application_name=application_name,
                        metrics=metrics,
                        start_time=start_time,
                        span=span,
                        capture_message_content=capture_message_content,
                        disable_metrics=disable_metrics,
                        version=version,
                        **kwargs,
                    )

                except Exception as e:
                    handle_exception(span, e)

                return response

    return wrapper


def aembedding(
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for GenAI embedding function call
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI embedding function call.
        """
        server_address, server_port = set_server_address_and_port(
            instance, "NOT_FOUND", "NOT_FOUND"
        )
        request_model = kwargs.get("model", "text-embedding-ada-002")

        span_name = (
            f"{SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING} {request_model}"
        )

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)

            try:
                response = process_embedding_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs,
                )

            except Exception as e:
                handle_exception(span, e)

            return response

    return wrapper



"""
Utilities for Qdrant instrumentation.
"""

import time
from urllib.parse import urlparse
from opentelemetry.trace import Status, StatusCode
from openlit.__helpers import (
    common_db_span_attributes,
    record_db_metrics,
)
from openlit.semcov import SemanticConvention

# Operation mapping for Qdrant endpoints
DB_OPERATION_MAP = {
    "qdrant.create_collection": SemanticConvention.DB_OPERATION_CREATE_COLLECTION,
    "qdrant.upload_collection": SemanticConvention.DB_OPERATION_CREATE_COLLECTION,
    "qdrant.delete_collection": SemanticConvention.DB_OPERATION_DELETE_COLLECTION,
    "qdrant.update_collection": SemanticConvention.DB_OPERATION_UPDATE_COLLECTION,
    "qdrant.upsert": SemanticConvention.DB_OPERATION_UPSERT,
    "qdrant.upload_points": SemanticConvention.DB_OPERATION_INSERT,
    "qdrant.set_payload": SemanticConvention.DB_OPERATION_INSERT,
    "qdrant.overwrite_payload": SemanticConvention.DB_OPERATION_UPDATE,
    "qdrant.update_vectors": SemanticConvention.DB_OPERATION_UPDATE,
    "qdrant.delete": SemanticConvention.DB_OPERATION_DELETE,
    "qdrant.delete_vectors": SemanticConvention.DB_OPERATION_DELETE,
    "qdrant.delete_payload": SemanticConvention.DB_OPERATION_DELETE,
    "qdrant.clear_payload": SemanticConvention.DB_OPERATION_DELETE,
    "qdrant.retrieve": SemanticConvention.DB_OPERATION_GET,
    "qdrant.scroll": SemanticConvention.DB_OPERATION_GET,
    "qdrant.search": SemanticConvention.DB_OPERATION_GET,
    "qdrant.search_groups": SemanticConvention.DB_OPERATION_GET,
    "qdrant.recommend": SemanticConvention.DB_OPERATION_GET,
    "qdrant.query_points": SemanticConvention.DB_OPERATION_GET,
    "qdrant.create_payload_index": SemanticConvention.DB_OPERATION_CREATE_INDEX,
}


def object_count(obj):
    """
    Counts Length of object if it exists, Else returns 0.
    """
    try:
        return len(obj)
    except (TypeError, AttributeError):
        return 0


def set_server_address_and_port(instance):
    """
    Extracts server address and port from Qdrant client instance.

    Args:
        instance: Qdrant client instance

    Returns:
        tuple: (server_address, server_port)
    """
    server_address = "localhost"
    server_port = 6333

    # Try to extract actual server info from Qdrant init_options
    if hasattr(instance, "init_options") and instance.init_options:
        init_options = instance.init_options

        # Get URL and extract host/port from it
        url = init_options.get("url", "")
        if url:
            try:
                parsed = urlparse(url)
                if parsed.hostname:
                    server_address = parsed.hostname
                if parsed.port:
                    server_port = parsed.port
            except Exception:
                pass

        # Also try direct port from init_options if URL parsing didnt work
        if "port" in init_options and init_options["port"]:
            server_port = init_options["port"]

    return server_address, server_port


def common_qdrant_logic(
    scope,
    environment,
    application_name,
    metrics,
    capture_message_content,
    disable_metrics,
    version,
    instance=None,
    endpoint=None,
):
    """
    Process Qdrant request and generate telemetry.
    """
    scope._end_time = time.time()

    # Set common database span attributes using helper
    common_db_span_attributes(
        scope,
        SemanticConvention.DB_SYSTEM_QDRANT,
        scope._server_address,
        scope._server_port,
        environment,
        application_name,
        version,
    )

    # Set DB operation specific attributes
    scope._span.set_attribute(SemanticConvention.DB_OPERATION_NAME, scope._db_operation)
    scope._span.set_attribute(
        SemanticConvention.DB_CLIENT_OPERATION_DURATION,
        scope._end_time - scope._start_time,
    )

    # Handle collection management operations
    if scope._db_operation in [
        SemanticConvention.DB_OPERATION_CREATE_COLLECTION,
        SemanticConvention.DB_OPERATION_DELETE_COLLECTION,
        SemanticConvention.DB_OPERATION_UPDATE_COLLECTION,
    ]:
        collection_name = scope._kwargs.get("collection_name", "unknown")

        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_TEXT, f"Collection: {collection_name}"
        )
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name}",
        )

    # Handle data insertion operations
    elif scope._db_operation in [
        SemanticConvention.DB_OPERATION_INSERT,
        SemanticConvention.DB_OPERATION_UPSERT,
    ]:
        collection_name = scope._kwargs.get("collection_name", "unknown")

        if endpoint == "qdrant.set_payload":
            points = scope._kwargs.get("points", [])
            payload = scope._kwargs.get("payload", {})

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(points))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points)
            )
            scope._span.set_attribute(
                SemanticConvention.DB_PAYLOAD_COUNT, object_count(payload)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"points={object_count(points)} "
                f"payload={object_count(payload)}",
            )

        elif endpoint in ["qdrant.upsert", "qdrant.upload_points"]:
            points = scope._kwargs.get("points", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(points))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"points={object_count(points)}",
            )

    # Handle data update operations
    elif scope._db_operation == SemanticConvention.DB_OPERATION_UPDATE:
        collection_name = scope._kwargs.get("collection_name", "unknown")

        if endpoint == "qdrant.overwrite_payload":
            points = scope._kwargs.get("points", [])
            payload = scope._kwargs.get("payload", {})

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(points))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points)
            )
            scope._span.set_attribute(
                SemanticConvention.DB_PAYLOAD_COUNT, object_count(payload)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"points={object_count(points)} "
                f"payload={object_count(payload)}",
            )

        elif endpoint == "qdrant.update_vectors":
            points = scope._kwargs.get("points", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(points))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"points={object_count(points)}",
            )

    # Handle data deletion operations
    elif scope._db_operation == SemanticConvention.DB_OPERATION_DELETE:
        collection_name = scope._kwargs.get("collection_name", "unknown")

        if endpoint in ["qdrant.delete_payload", "qdrant.delete_vectors"]:
            points = scope._kwargs.get("points", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(points))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"points={object_count(points)}",
            )

        elif endpoint in ["qdrant.clear_payload", "qdrant.delete"]:
            points_selector = scope._kwargs.get("points_selector", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_TEXT, str(points_selector)
            )
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(points_selector)
            )

            # Set operation status if response available
            if scope._response and hasattr(scope._response, "status"):
                scope._span.set_attribute(
                    SemanticConvention.DB_OPERATION_STATUS, scope._response.status
                )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"selector={object_count(points_selector)}",
            )

    # Handle query operations
    elif scope._db_operation == SemanticConvention.DB_OPERATION_GET:
        collection_name = scope._kwargs.get("collection_name", "unknown")

        if endpoint == "qdrant.retrieve":
            ids = scope._kwargs.get("ids", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(ids))
            scope._span.set_attribute(
                SemanticConvention.DB_VECTOR_COUNT, object_count(ids)
            )

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} ids={ids}",
            )

        elif endpoint == "qdrant.scroll":
            scroll_filter = scope._kwargs.get("scroll_filter", {})

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_TEXT, str(scroll_filter)
            )
            scope._span.set_attribute(SemanticConvention.DB_FILTER, str(scroll_filter))

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} filter={scroll_filter}",
            )

        elif endpoint in ["qdrant.search", "qdrant.search_groups"]:
            query_vector = scope._kwargs.get("query_vector", [])
            limit = scope._kwargs.get("limit", 10)

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_TEXT, str(query_vector)
            )
            scope._span.set_attribute(SemanticConvention.DB_VECTOR_QUERY_TOP_K, limit)

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} limit={limit}",
            )

        elif endpoint == "qdrant.recommend":
            positive = scope._kwargs.get("positive", [])
            negative = scope._kwargs.get("negative", [])

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            query_content = f"positive:{positive} negative:{negative}"
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, query_content)

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} "
                f"positive={object_count(positive)} "
                f"negative={object_count(negative)}",
            )

        elif endpoint == "qdrant.query_points":
            query = scope._kwargs.get("query", {})

            scope._span.set_attribute(
                SemanticConvention.DB_COLLECTION_NAME, collection_name
            )
            scope._span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(query))

            scope._span.set_attribute(
                SemanticConvention.DB_QUERY_SUMMARY,
                f"{scope._db_operation} {collection_name} query={query}",
            )

    # Handle index operations
    elif scope._db_operation == SemanticConvention.DB_OPERATION_CREATE_INDEX:
        collection_name = scope._kwargs.get("collection_name", "unknown")
        field_name = scope._kwargs.get("field_name", "unknown")

        scope._span.set_attribute(
            SemanticConvention.DB_COLLECTION_NAME, collection_name
        )
        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_TEXT, f"Field: {field_name}"
        )

        scope._span.set_attribute(
            SemanticConvention.DB_QUERY_SUMMARY,
            f"{scope._db_operation} {collection_name} field={field_name}",
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Record metrics using helper
    if not disable_metrics:
        record_db_metrics(
            metrics,
            SemanticConvention.DB_SYSTEM_QDRANT,
            scope._server_address,
            scope._server_port,
            environment,
            application_name,
            scope._start_time,
            scope._end_time,
        )


def process_qdrant_response(
    response,
    db_operation,
    server_address,
    server_port,
    environment,
    application_name,
    metrics,
    start_time,
    span,
    capture_message_content,
    disable_metrics,
    version,
    instance=None,
    args=None,
    endpoint=None,
    **kwargs,
):
    """
    Process Qdrant response and generate telemetry.
    """
    # Create scope object
    scope = type("GenericScope", (), {})()
    scope._span = span
    scope._response = response
    scope._db_operation = db_operation
    scope._server_address = server_address
    scope._server_port = server_port
    scope._start_time = start_time
    scope._kwargs = kwargs

    # Process the response
    common_qdrant_logic(
        scope,
        environment,
        application_name,
        metrics,
        capture_message_content,
        disable_metrics,
        version,
        instance=instance,
        endpoint=endpoint,
    )

    return response



# pylint: disable=duplicate-code, broad-exception-caught, too-many-statements, unused-argument, too-many-branches
"""
Module for monitoring Letta calls following OpenTelemetry patterns.
"""

import logging
import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import handle_exception
from .utils import (
    get_span_name,
    process_letta_response,
    TracedLettaStream,
    OPERATION_TYPE_MAP,
)

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)


def create_letta_wrapper(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """
    Creates a unified telemetry wrapper for Letta operations following OpenTelemetry patterns.

    This follows the same pattern as LiteLLM and CrewAI instrumentations.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the API call to add telemetry following OpenTelemetry semantic conventions.
        """

        # Extract operation type from endpoint
        endpoint_parts = gen_ai_endpoint.split(".")
        operation = endpoint_parts[-1] if endpoint_parts else "unknown"
        operation_type = OPERATION_TYPE_MAP.get(operation, "workflow")

        # Generate proper span name with context
        span_name = get_span_name(operation_type, gen_ai_endpoint, instance, kwargs)
        start_time = time.time()

        # Check if this is a streaming operation
        streaming = operation_type == "chat"  # Assume chat operations might stream

        if streaming:
            # Execute the operation first to check if it's actually streaming
            response = wrapped(*args, **kwargs)

            # Check if response is actually a stream
            if hasattr(response, "__iter__") and hasattr(response, "__next__"):
                # For streaming: create span without context manager (like LiteLLM)
                span = tracer.start_span(span_name, kind=SpanKind.CLIENT)
                return TracedLettaStream(
                    wrapped_stream=response,
                    span=span,
                    span_name=span_name,
                    kwargs=kwargs,
                    operation_type=operation_type,
                    instance=instance,
                    start_time=start_time,
                    environment=environment,
                    application_name=application_name,
                    version=version,
                    endpoint=gen_ai_endpoint,
                    capture_content=capture_message_content,
                    pricing_info=pricing_info,
                    tracer=tracer,
                )

        # Non-streaming or non-chat operations
        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            try:
                if not streaming:
                    # Execute the operation for non-streaming
                    response = wrapped(*args, **kwargs)

                # Process non-streaming response using common helpers
                process_letta_response(
                    span,
                    response,
                    kwargs,
                    operation_type,
                    instance,
                    start_time,
                    environment,
                    application_name,
                    version,
                    gen_ai_endpoint,
                    capture_message_content,
                    pricing_info,
                )

                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in Letta trace creation: %s", e)
                # Re-execute without instrumentation on error
                return wrapped(*args, **kwargs)

    return wrapper


# Agent operations
def create_agent(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Agent operations wrapper (create, retrieve, modify, delete, list)"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Message operations
def send_message(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Message operations wrapper (create, create_stream, list, modify, etc.)"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Memory operations
def memory_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Memory operations wrapper (core memory, blocks)"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Tool operations
def tool_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Tool operations wrapper (list, attach, detach, create)"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Context operations
def context_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Context operations wrapper"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )


# Source operations
def source_operation(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
    metrics,
    disable_metrics,
):
    """Source operations wrapper"""
    return create_letta_wrapper(
        gen_ai_endpoint,
        version,
        environment,
        application_name,
        tracer,
        pricing_info,
        capture_message_content,
        metrics,
        disable_metrics,
    )



# pylint: disable=duplicate-code, broad-exception-caught, too-many-statements, unused-argument
"""
Optimized sync mem0 instrumentation following OpenLIT Framework Guide.
"""

import time
from opentelemetry.trace import SpanKind, Status, StatusCode, set_span_in_context
from opentelemetry import context
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from openlit.__helpers import handle_exception
from openlit.instrumentation.mem0.utils import (
    patch_concurrent_futures_context,
    Mem0Context,
    set_mem0_span_attributes,
    set_mem0_content_attributes,
    logger,
)


def mem0_wrap(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    capture_message_content,
):
    """
    Optimized wrapper for sync mem0 operations with proper hierarchy and performance.

    Returns:
        Wrapper function for mem0 operations
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Inner wrapper with comprehensive tracing and optimized performance.
        """
        # Suppression check
        if context.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            return wrapped(*args, **kwargs)

        # Handle Memory.__init__ to create parent span for initialization operations
        if gen_ai_endpoint == "memory init":
            with tracer.start_as_current_span("memory init", kind=SpanKind.INTERNAL):
                return wrapped(*args, **kwargs)

        # Determine span type efficiently
        is_internal = any(
            marker in gen_ai_endpoint
            for marker in ["_", "vector_store", "graph", "create_memory", "init"]
        )
        span_kind = SpanKind.INTERNAL if is_internal else SpanKind.CLIENT

        # Create optimized context
        ctx = Mem0Context(
            instance, args, kwargs, version, environment, application_name
        )
        current_context = context.get_current()

        with tracer.start_as_current_span(
            gen_ai_endpoint, kind=span_kind, context=current_context
        ) as span:
            span_context = set_span_in_context(span, context=current_context)
            token = context.attach(span_context)

            start_time = time.perf_counter()

            try:
                # Apply threading context fix for all CLIENT operations (mem0 operations use ThreadPoolExecutor)
                if span_kind == SpanKind.CLIENT:
                    restore_patch = patch_concurrent_futures_context(span_context)
                    try:
                        response = wrapped(*args, **kwargs)
                    finally:
                        restore_patch()
                else:
                    response = wrapped(*args, **kwargs)

                # Calculate duration
                duration = time.perf_counter() - start_time

                # Set span attributes
                set_mem0_span_attributes(span, gen_ai_endpoint, ctx, response)
                span.set_attribute("gen_ai.client.operation.duration", duration)

                # Set content attributes if enabled
                set_mem0_content_attributes(
                    span, gen_ai_endpoint, ctx, response, capture_message_content
                )

                span.set_status(Status(StatusCode.OK))
                return response

            except Exception as e:
                duration = time.perf_counter() - start_time
                span.set_attribute("gen_ai.client.operation.duration", duration)

                try:
                    set_mem0_span_attributes(span, gen_ai_endpoint, ctx)
                except Exception as attr_error:
                    logger.debug(
                        "Failed to set span attributes on error: %s", attr_error
                    )

                handle_exception(span, e)
                raise

            finally:
                context.detach(token)

    return wrapper
